import os
import numpy as np
import pandas as pd


class Energy_Systems:
    """ A class of methods to initialise energy sytems and create the data/model files needed for optimisation.
    """
    def __init__(self, year, region, emission, technology, capacity_technology,
                 availability_technology, fuel, specified_fuel,
                 accumulated_fuel, timeslice, mode_of_operation, storage,
                 daytype, season, dailytimebracket):
        """ Function to create complete energy system set to prepare datafile, as per the established model.

        Args:
            Sets:
                year (list): Set of Years
                region (region): Set of Regions
                emission (list): Set of Emissions
        """
        self.year = year
        self.region = region
        self.emission = emission
        self.technology = technology
        self.capacity_technology = capacity_technology
        self.availability_technology = availability_technology
        self.fuel = fuel
        self.specified_fuel = specified_fuel
        self.accumulated_fuel = accumulated_fuel
        self.timeslice = timeslice
        self.mode_of_operation = mode_of_operation
        self.storage = storage
        self.daytype = daytype
        self.season = season
        self.dailytimebracket = dailytimebracket

        ly = len(self.year)
        lr = len(self.region)
        le = len(self.emission)
        lt = len(self.technology)
        lct = len(self.capacity_technology)
        lat = len(self.availability_technology)
        lf = len(self.fuel)
        lsf = len(self.specified_fuel)
        laf = len(self.specified_fuel)
        ll = len(self.timeslice)
        lm = len(self.mode_of_operation)
        ls = len(self.storage)
        lld = len(self.daytype)
        lls = len(self.season)
        llh = len(self.dailytimebracket)

        self.ly = ly
        self.lr = lr
        self.le = le
        self.lt = lt
        self.lct = lct
        self.lat = lat
        self.lf = lf
        self.lsf = lsf
        self.laf = laf
        self.ll = ll
        self.lm = lm
        self.ls = ls
        self.lld = lld
        self.lls = lls
        self.llh = llh

        self.YearSplit = np.ones((ll, ly))
        self.DiscountRate = np.ones((lr))
        self.DaySplit = np.ones((llh, ly))
        self.Conversionls = np.ones((ll, lls))
        self.Conversionld = np.ones((ll, lld))
        self.Conversionlh = np.ones((ll, llh))
        self.DaysInDayType = np.ones((lls, lld, ly))
        self.TradeRoute = np.ones((lr, lr, lf, ly))
        self.DepreciationMethod = np.ones((lr))
        self.SpecifiedAnnualDemand = np.ones((lr, lsf, ly))
        self.SpecifiedDemandProfile = np.ones((lr, lf, ll, ly))
        self.AccumulatedAnnualDemand = np.ones((lr, laf, ly))
        self.CapacityToActivityUnit = np.ones((lr, lct))
        self.CapacityFactor = np.ones((lr, lt, ll, ly))
        self.AvailabilityFactor = np.ones((lr, lt, ly))
        self.OperationalLife = np.ones((lr, lt))
        self.ResidualCapacity = np.ones((lr, lt, ly))
        self.InputActivityRatio = np.ones((lr, lt, lf, lm, ly))
        self.OutputActivityRatio = np.ones((lr, lt, lf, lm, ly))
        self.CapitalCost = np.ones((lr, lt, ly))
        self.VariableCost = np.ones((lr, lt, lm, ly))
        self.FixedCost = np.ones((lr, lt, ly))
        self.TechnologyToStorage = np.ones((lr, lt, ls, lm))
        self.TechnologyFromStorage = np.ones((lr, lt, ls, lm))
        self.StorageLevelStart = np.ones((lr, ls))
        self.StorageMaxChargeRate = np.ones((lr, ls))
        self.StorageMaxDischargeRate = np.ones((lr, ls))
        self.MinStorageCharge = np.ones((lr, ls, ly))
        self.OperationalLifeStorage = np.ones((lr, ls))
        self.CapitalCostStorage = np.ones((lr, ls, ly))
        self.ResidualStorageCapacity = np.ones((lr, ls, ly))
        self.CapacityOfOneTechnologyUnit = np.ones((lr, lt, ly))
        self.TotalAnnualMaxCapacity = np.ones((lr, lt, ly))
        self.TotalAnnualMinCapacity = np.ones((lr, lt, ly))
        self.TotalAnnualMaxCapacityInvestment = np.ones((lr, lt, ly))
        self.TotalAnnualMinCapacityInvestment = np.ones((lr, lt, ly))
        self.TotalTechnologyAnnualActivityLowerLimit = np.ones((lr, lt, ly))
        self.TotalTechnologyAnnualActivityUpperLimit = np.ones((lr, lt, ly))
        self.TotalTechnologyModelPeriodActivityUpperLimit = np.ones((lr, lt))
        self.TotalTechnologyModelPeriodActivityLowerLimit = np.ones((lr, lt))
        self.ReserveMarginTagTechnology = np.ones((lr, lt, ly))
        self.ReserveMarginTagFuel = np.ones((lr, lf, ly))
        self.ReserveMargin = np.ones((lr, ly))
        self.RETagTechnology = np.ones((lr, lt, ly))
        self.RETagFuel = np.ones((lr, lf, ly))
        self.REMinProductionTarget = np.ones((lr, ly))
        self.EmissionActivityRatio = np.ones((lr, lt, le, lm, ly))
        self.EmissionsPenalty = np.ones((lr, le, ly))
        self.AnnualExogenousEmission = np.ones((lr, le, ly))
        self.AnnualEmissionLimit = np.ones((lr, le, ly))
        self.ModelPeriodExogenousEmission = np.ones((lr, le))
        self.ModelPeriodEmissionLimit = np.ones((lr, le))

    def load_datacase(self, case, system):
        """ Loads the data case to a correct configured and intialised energy system
            (The load status dictionary must be compatible with the data_case and system_case)

        Args:
            case (object): Energy system datacase
            system (object): Initialised energy system
            load_status (dict): Dictionary setting the required sets and parameters to load

        Returns:
            system_case (dict): Returns the updated dictionary
        """
        # Loads the sets to the energy system
        system.year = case.year
        system.region = case.region
        system.emission = case.emission
        system.technology = case.technology
        system.fuel = case.fuel
        system.timeslice = case.timeslice
        system.mode_of_operation = case.mode_of_operation
        system.storage = case.storage
        system.daytype = case.daytype
        system.season = case.season
        system.dailytimebracket = case.dailytimebracket
        # Loads the parameters to the energy system
        system.YearSplit = case.YearSplit
        system.DiscountRate = case.DiscountRate
        system.DaySplit = case.DaySplit
        system.Conversionls = case.Conversionls
        system.Conversionld = case.Conversionld
        system.Conversionlh = case.Conversionlh
        system.DaysInDayType = case.DaysInDayType
        system.TradeRoute = case.TradeRoute
        system.DepreciationMethod = case.DepreciationMethod
        system.SpecifiedAnnualDemand = case.SpecifiedAnnualDemand
        system.SpecifiedDemandProfile = case.SpecifiedDemandProfile
        system.AccumulatedAnnualDemand = case.AccumulatedAnnualDemand
        system.CapacityToActivityUnit = case.CapacityToActivityUnit
        system.CapacityFactor = case.CapacityFactor
        system.AvailabilityFactor = case.AvailabilityFactor
        system.OperationalLife = case.OperationalLife
        system.ResidualCapacity = case.ResidualCapacity
        system.InputActivityRatio = case.InputActivityRatio
        system.OutputActivityRatio = case.OutputActivityRatio
        system.CapitalCost = case.CapitalCost
        system.VariableCost = case.VariableCost
        system.FixedCost = case.FixedCost
        system.TechnologyToStorage = case.TechnologyToStorage
        system.TechnologyFromStorage = case.TechnologyFromStorage
        system.StorageLevelStart = case.StorageLevelStart
        system.StorageMaxChargeRate = case.StorageMaxChargeRate
        system.StorageMaxDischargeRate = case.StorageMaxDischargeRate
        system.MinStorageCharge = case.MinStorageCharge
        system.OperationalLifeStorage = case.OperationalLifeStorage
        system.CapitalCostStorage = case.CapitalCostStorage
        system.ResidualStorageCapacity = case.ResidualStorageCapacity
        system.CapacityOfOneTechnologyUnit = case.CapacityOfOneTechnologyUnit
        system.TotalAnnualMaxCapacity = case.TotalAnnualMaxCapacity
        system.TotalAnnualMinCapacity = case.TotalAnnualMinCapacity
        system.TotalAnnualMaxCapacityInvestment = case.TotalAnnualMaxCapacityInvestment
        system.TotalAnnualMinCapacityInvestment = case.TotalAnnualMinCapacityInvestment
        system.TotalTechnologyAnnualActivityLowerLimit = case.TotalTechnologyAnnualActivityLowerLimit
        system.TotalTechnologyAnnualActivityUpperLimit = case.TotalTechnologyAnnualActivityUpperLimit
        system.TotalTechnologyModelPeriodActivityUpperLimit = case.TotalTechnologyModelPeriodActivityUpperLimit
        system.TotalTechnologyModelPeriodActivityLowerLimit = case.TotalTechnologyModelPeriodActivityLowerLimit
        system.ReserveMarginTagTechnology = case.ReserveMarginTagTechnology
        system.ReserveMarginTagFuel = case.ReserveMarginTagFuel
        system.ReserveMargin = case.ReserveMargin
        system.RETagTechnology = case.RETagTechnology
        system.RETagFuel = case.RETagFuel
        system.REMinProductionTarget = case.REMinProductionTarget
        system.EmissionActivityRatio = case.EmissionActivityRatio
        system.EmissionsPenalty = case.EmissionsPenalty
        system.AnnualExogenousEmission = case.AnnualExogenousEmission
        system.AnnualEmissionLimit = case.AnnualEmissionLimit
        system.ModelPeriodExogenousEmission = case.ModelPeriodExogenousEmission
        system.ModelPeriodEmissionLimit = case.ModelPeriodEmissionLimit

    def create_model_file(self, root, file):
        """Creates the model file necessary for the project to run
        
        Args: 
            Parameters for the basic problem
        
        Returns: 
            The loaded in parameters and sets
    
        """
        # Finds the file
        # data = Find_File(data_file,model_file)
        model_location = os.path.join(root, file)
        df = pd.read_excel(model_location, sheet_name='Model')
        # Creates a new dataframe based on the variables on the Include column values
        df_Include = df[df.Include == 'Yes']
        df_model = df_Include[['Name']].copy()

        # Creates a file location and write the model to a text file
        model_txt = 'GOCPI_OseMOSYS_Model.txt'
        model_location = os.path.join(root, model_txt)

        # Saves the user defined model to a text file
        np.savetxt(model_location, df_model.values, fmt='%s')

    def create_data_file(self, file_location, defaults_dictionary,
                         toggle_defaults):
        """ Creates the osemosys datafile

        Args:
            file_location (str): String of directory to save data file
            defaults_dictionary (dict): Dictionary setting the default values for parameters
            toggle_defaults (Bool): Boolean (True/False to only print the default functions
        """
        # Opens the file for write the data
        with open(file_location, 'w') as f:
            # Sets up the preamble for the data file
            f.write('# GOCPI Energy System Data File\n')
            f.write(
                '# Insert instructions when the file is running properly\n')
            f.write('#\n')
            # Sets
            f.write('# Sets\n#\n')
            # year
            set_string = ' '.join(self.year)
            f.write('set YEAR\t:=\t{0};\n'.format(set_string))
            # region
            set_string = ' '.join(self.region)
            f.write('set REGION\t:=\t{0};\n'.format(set_string))
            # emission
            set_string = ' '.join(self.emission)
            f.write('set EMISSION\t:=\t{0};\n'.format(set_string))
            # technology
            set_string = ' '.join(self.technology)
            f.write('set TECHNOLOGY\t:=\t{0};\n'.format(set_string))
            # fuel
            set_string = ' '.join(self.fuel)
            f.write('set FUEL\t:=\t{0};\n'.format(set_string))
            # timeslice
            set_string = ' '.join(self.timeslice)
            f.write('set TIMESLICE\t:=\t{0};\n'.format(set_string))
            # mode_of_operation
            set_string = ' '.join(self.mode_of_operation)
            f.write('set MODE_OF_OPERATION\t:=\t{0};\n'.format(set_string))
            # storage
            set_string = ' '.join(self.storage)
            f.write('set STORAGE\t:=\t{0};\n'.format(set_string))
            # daytype
            set_string = ' '.join(self.daytype)
            f.write('set DAYTYPE\t:=\t{0};\n'.format(set_string))
            # season
            set_string = ' '.join(self.season)
            f.write('set SEASON\t:=\t{0};\n'.format(set_string))
            # dailytimebracket
            set_string = ' '.join(self.dailytimebracket)
            f.write('set DAILYTIMEBRACKET\t:=\t{0};\n'.format(set_string))
            f.write('#\n')
            # Parameters

            # YearSplit = np.zeros((ll,ly))
            param = 'YearSplit'
            f.write('#\n')
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            columns = self.year
            column_string = ' '.join(columns)
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            if toggle_defaults[param] == True:
                # Converts maxtrix rows to list
                array = np.array(self.timeslice)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.YearSplit[:, :]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # DiscountRate = np.zeros((lr))
            param = 'DiscountRate'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Converts maxtrix rows to list
            array = np.array(self.region)
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.DiscountRate[:]
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # DaySplit = np.zeros((llh,ly))
            param = 'DaySplit'
            f.write('#\n')
            columns = self.year
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.dailytimebracket)
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.DaySplit[:, :]
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # Conversionls = np.zeros((ll,ls))
            param = 'Conversionls'  # Change this line
            f.write('#\n')
            columns = self.season  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.timeslice)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.Conversionls[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # Conversionld = np.zeros((ll,lld))
            param = 'Conversionld'  # Change this line
            f.write('#\n')
            columns = self.daytype  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.timeslice)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.Conversionld[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # Conversionlh = np.zeros((ll,llh))
            param = 'Conversionlh'  # Change this line
            f.write('#\n')
            columns = self.dailytimebracket  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.timeslice)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.Conversionlh[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # DaysInDayType = np.zeros((lls,lld,ly))
            #Writes new line character at parameter metadata to the text file
            param = 'DaysInDayType'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.daytype
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.season)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.DaysInDayType[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TradeRoute = np.zeros((lr,lr,lf,ly))
            param = 'TradeRoute'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.lf):
                # Sets index value for format string
                fl = self.fuel[j]
                for k in range(self.ly):
                    # Sets index value for format string
                    y = self.year[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.region
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.TradeRoute[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        fl, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')
            # DepreciationMethod = np.zeros((lr))
            param = 'DepreciationMethod'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Converts maxtrix rows to list
            array = np.array(self.region)
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.DepreciationMethod[:]
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # SpecifiedAnnualDemand = np.zeros((lr,lsf,ly))
            param = 'SpecifiedAnnualDemand'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.specified_fuel
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.SpecifiedAnnualDemand[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # SpecifiedDemandProfile = np.zeros((lr,lf,ll,ly))
            param = 'SpecifiedDemandProfile'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.ll):
                # Sets index value for format string
                x = self.timeslice[j]
                for k in range(self.ly):
                    # Sets index value for format string
                    y = self.year[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.specified_fuel
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.SpecifiedDemandProfile[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        x, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # AccumulatedAnnualDemand = np.zeros((lr,lf,ly))
            param = 'AccumulatedAnnualDemand'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.accumulated_fuel
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.AccumulatedAnnualDemand[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # CapacityToActivityUnit = np.zeros((lr,lt))
            param = 'CapacityToActivityUnit'  # Change this line
            f.write('#\n')
            columns = self.capacity_technology  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.CapacityToActivityUnit[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # CapacityFactor = np.zeros((lr,lt,ll,ly))
            param = 'CapacityFactor'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.ll):
                # Sets index value for format string
                x = self.timeslice[j]
                for k in range(self.ly):
                    # Sets index value for format string
                    y = self.year[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.capacity_technology
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.CapacityFactor[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        x, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # AvailabilityFactor = np.zeros((lr,lt,ly))
            param = 'AvailabilityFactor'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.availability_technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.AvailabilityFactor[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # OperationalLife = np.zeros((lr,lt))
            param = 'OperationalLife'  # Change this line
            f.write('#\n')

            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            if self.OperationalLife.all() != None:
                columns = self.technology  # Change this line
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)  # Change this line
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.OperationalLife[:, :]  # Change this line
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ResidualCapacity = np.zeros((lr,lt,ly))
            param = 'ResidualCapacity'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.ResidualCapacity[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # InputActivityRatio = np.zeros((lr,lt,lf,lm,ly))
            param = 'InputActivityRatio'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for i in range(self.lf):  # Change loops if you need
                # Sets index value for format string
                x = self.fuel[i]
                for j in range(self.lm):
                    # Sets index value for format string
                    y = self.mode_of_operation[j]
                    for k in range(self.ly):
                        # Sets index value for format string
                        z = self.year[k]
                        # Converts matrix columns to strings columns to strings
                        columns = self.technology
                        column_string = ' '.join(columns)
                        # Converts maxtrix rows to list
                        array = np.array(self.region)
                        array = array.T
                        lt = array.tolist()
                        # Creates 2D matrix for this value
                        mat = self.InputActivityRatio[:, :, i, j, k]
                        # Converts combined matrix to list and combines lists
                        matlist = mat.tolist()
                        #Combines the two lists
                        combined_list = list(zip(lt, matlist))
                        # Writes index specific parameter values to the text files
                        f.write("\t[*,*,{0},{1},{2}]:\t{3}\t:=\n".format(
                            x, y, z, column_string))
                        for line in combined_list:
                            combinedflat = ''.join(str(line))
                            combinedflat = combinedflat.replace('[', '')
                            combinedflat = combinedflat.replace(']', '')
                            combinedflat = combinedflat.replace("'", '')
                            combinedflat = combinedflat.replace(",", '')
                            combinedflat = combinedflat.replace("(", '')
                            combinedflat = combinedflat.replace(")", '')
                            f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # OutputActivityRatio = np.zeros((lr,lt,lf,lm,ly))
            param = 'OutputActivityRatio'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for i in range(self.lf):  # Change loops if you need
                # Sets index value for format string
                x = self.fuel[i]
                for j in range(self.lm):
                    # Sets index value for format string
                    y = self.mode_of_operation[j]
                    for k in range(self.ly):
                        # Sets index value for format string
                        z = self.year[k]
                        # Converts matrix columns to strings columns to strings
                        columns = self.technology
                        column_string = ' '.join(columns)
                        # Converts maxtrix rows to list
                        array = np.array(self.region)
                        array = array.T
                        lt = array.tolist()
                        # Creates 2D matrix for this value
                        mat = self.OutputActivityRatio[:, :, i, j, k]
                        # Converts combined matrix to list and combines lists
                        matlist = mat.tolist()
                        #Combines the two lists
                        combined_list = list(zip(lt, matlist))
                        # Writes index specific parameter values to the text files
                        f.write("\t[*,*,{0},{1},{2}]:\t{3}\t:=\n".format(
                            x, y, z, column_string))
                        for line in combined_list:
                            combinedflat = ''.join(str(line))
                            combinedflat = combinedflat.replace('[', '')
                            combinedflat = combinedflat.replace(']', '')
                            combinedflat = combinedflat.replace("'", '')
                            combinedflat = combinedflat.replace(",", '')
                            combinedflat = combinedflat.replace("(", '')
                            combinedflat = combinedflat.replace(")", '')
                            f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # CapitalCost = np.zeros((lr,lt,ly))
            param = 'CapitalCost'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.CapitalCost[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # VariableCost = np.zeros((lr,lt,lm,ly))
            param = 'VariableCost'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.lm):
                # Sets index value for format string
                x = self.mode_of_operation[j]
                for k in range(self.ly):
                    # Sets index value for format string
                    y = self.year[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.technology
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.VariableCost[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        x, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # FixedCost = np.zeros((lr,lt,ly))
            param = 'FixedCost'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.FixedCost[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TechnologyToStorage = np.zeros((lr,lt,ls,lm))
            param = 'TechnologyToStorage'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.ls):
                # Sets index value for format string
                x = self.storage[j]
                for k in range(self.lm):
                    # Sets index value for format string
                    y = self.mode_of_operation[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.technology
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.TechnologyToStorage[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        x, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TechnologyFromStorage = np.zeros((lr,lt,ls,lm))
            param = 'TechnologyFromStorage'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for j in range(self.ls):
                # Sets index value for format string
                x = self.storage[j]
                for k in range(self.lm):
                    # Sets index value for format string
                    y = self.mode_of_operation[k]
                    # Converts matrix columns to strings columns to strings
                    columns = self.technology
                    column_string = ' '.join(columns)
                    # Converts maxtrix rows to list
                    array = np.array(self.region)
                    array = array.T
                    lt = array.tolist()
                    # Creates 2D matrix for this value
                    mat = self.TechnologyFromStorage[:, :, j, k]
                    # Converts combined matrix to list and combines lists
                    matlist = mat.tolist()
                    #Combines the two lists
                    combined_list = list(zip(lt, matlist))
                    # Writes index specific parameter values to the text files
                    f.write("\t[*,*,{0},{1}]:\t{2}\t:=\n".format(
                        x, y, column_string))
                    for line in combined_list:
                        combinedflat = ''.join(str(line))
                        combinedflat = combinedflat.replace('[', '')
                        combinedflat = combinedflat.replace(']', '')
                        combinedflat = combinedflat.replace("'", '')
                        combinedflat = combinedflat.replace(",", '')
                        combinedflat = combinedflat.replace("(", '')
                        combinedflat = combinedflat.replace(")", '')
                        f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # StorageLevelStart = np.zeros((lr,ls))
            param = 'StorageLevelStart'  # Change this line
            f.write('#\n')
            columns = self.storage  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.StorageLevelStart[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # StorageMaxChargeRate = np.zeros((lr,ls))
            param = 'StorageMaxChargeRate'  # Change this line
            f.write('#\n')
            columns = self.storage  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.StorageMaxChargeRate[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # StorageMaxDischargeRate = np.zeros((lr,ls))
            param = 'StorageMaxDischargeRate'  # Change this line
            f.write('#\n')
            columns = self.storage  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.StorageMaxDischargeRate[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # MinStorageCharge = np.zeros((lr,ls,ly))
            param = 'MinStorageCharge'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.storage
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.MinStorageCharge[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # OperationalLifeStorage = np.zeros((lr,ls))
            param = 'OperationalLifeStorage'  # Change this line
            f.write('#\n')
            columns = self.storage  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.OperationalLifeStorage[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # CapitalCostStorage = np.zeros((lr,ls,ly))
            param = 'CapitalCostStorage'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.storage
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.CapitalCostStorage[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ResidualStorageCapacity = np.zeros((lr,ls,ly))
            param = 'ResidualStorageCapacity'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.storage
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.ResidualStorageCapacity[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # CapacityOfOneTechnologyUnit = np.zeros((lr,lt,ly))
            param = 'CapacityOfOneTechnologyUnit'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.CapacityOfOneTechnologyUnit[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalAnnualMaxCapacity = np.zeros((lr,lt,ly))
            param = 'TotalAnnualMaxCapacity'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalAnnualMaxCapacity[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalAnnualMinCapacity = np.zeros((lr,lt,ly))
            param = 'TotalAnnualMinCapacity'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalAnnualMinCapacity[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalAnnualMaxCapacityInvestment = np.zeros((lr,lt,ly))
            param = 'TotalAnnualMaxCapacityInvestment'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalAnnualMaxCapacityInvestment[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalAnnualMinCapacityInvestment = np.zeros((lr,lt,ly))
            param = 'TotalAnnualMinCapacityInvestment'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalAnnualMinCapacityInvestment[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalTechnologyAnnualActivityLowerLimit= np.zeros((lr,lt,ly))
            param = 'TotalTechnologyAnnualActivityLowerLimit'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalTechnologyAnnualActivityLowerLimit[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalTechnologyAnnualActivityUpperLimit = np.zeros((lr,lt,ly))
            param = 'TotalTechnologyAnnualActivityUpperLimit'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.TotalTechnologyAnnualActivityUpperLimit[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalTechnologyModelPeriodActivityUpperLimit = np.zeros((lr,lt))
            param = 'TotalTechnologyModelPeriodActivityUpperLimit'  # Change this line
            f.write('#\n')
            columns = self.technology  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.TotalTechnologyModelPeriodActivityUpperLimit[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # TotalTechnologyModelPeriodActivityLowerLimit = np.zeros((lr,lt))
            param = 'TotalTechnologyModelPeriodActivityLowerLimit'  # Change this line
            f.write('#\n')
            columns = self.technology  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.TotalTechnologyModelPeriodActivityLowerLimit[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ReserveMarginTagTechnology = np.zeros((lr,lt,ly))
            param = 'ReserveMarginTagTechnology'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.ReserveMarginTagTechnology[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ReserveMarginTagFuel = np.zeros((lr,lf,ly))
            param = 'ReserveMarginTagFuel'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.fuel
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.ReserveMarginTagFuel[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ReserveMargin = np.zeros((lr,ly))
            param = 'ReserveMargin'  # Change this line
            f.write('#\n')
            columns = self.year  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.ReserveMargin[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # RETagTechnology = np.zeros((lr,lt,ly))
            param = 'RETagTechnology'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.technology
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.RETagTechnology[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # RETagFuel = np.zeros((lr,lf,ly))
            param = 'RETagFuel'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.fuel
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.RETagFuel[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # REMinProductionTarget = np.zeros((lr,ly))
            param = 'REMinProductionTarget'  # Change this line
            f.write('#\n')
            columns = self.year  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.REMinProductionTarget[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # EmissionActivityRatio = np.zeros((lr,lt,le,lm,ly))
            #Writes new line character at parameter metadata to the text file
            param = 'EmissionActivityRatio'  # Change this line
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for i in range(self.le):  # Change loops if you need
                # Sets index value for format string
                emission = self.emission[i]
                for j in range(self.lm):
                    # Sets index value for format string
                    MOO = self.mode_of_operation[j]
                    for k in range(self.ly):
                        # Sets index value for format string
                        y = self.year[k]
                        # Converts matrix columns to strings columns to strings
                        columns = self.technology
                        column_string = ' '.join(columns)
                        # Converts maxtrix rows to list
                        array = np.array(self.region)
                        array = array.T
                        lt = array.tolist()
                        # Creates 2D matrix for this value
                        mat = self.EmissionActivityRatio[:, :, i, j, k]
                        # Converts combined matrix to list and combines lists
                        matlist = mat.tolist()
                        #Combines the two lists
                        combined_list = list(zip(lt, matlist))
                        # Writes index specific parameter values to the text files
                        f.write("\t[*,*,{0},{1},{2}]:\t{3}\t:=\n".format(
                            emission, MOO, y, column_string))
                        for line in combined_list:
                            combinedflat = ''.join(str(line))
                            combinedflat = combinedflat.replace('[', '')
                            combinedflat = combinedflat.replace(']', '')
                            combinedflat = combinedflat.replace("'", '')
                            combinedflat = combinedflat.replace(",", '')
                            combinedflat = combinedflat.replace("(", '')
                            combinedflat = combinedflat.replace(")", '')
                            f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # EmissionsPenalty = np.zeros((lr,le,ly))
            param = 'EmissionsPenalty'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.emission
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.EmissionsPenalty[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # AnnualExogenousEmission = np.zeros((lr,le,ly))
            param = 'AnnualExogenousEmission'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.emission
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.AnnualExogenousEmission[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # AnnualEmissionLimit = np.zeros((lr,le,ly))
            param = 'AnnualEmissionLimit'
            f.write('#\n')
            f.write("param\t{0}\tdefault\t{1}:=\n".format(
                param, defaults_dictionary[param]))
            # Writes parameter values to the text files
            for k in range(self.ly):
                # Sets index value for format string
                y = self.year[k]
                # Converts matrix columns to strings columns to strings
                columns = self.emission
                column_string = ' '.join(columns)
                # Converts maxtrix rows to list
                array = np.array(self.region)
                array = array.T
                lt = array.tolist()
                # Creates 2D matrix for this value
                mat = self.AnnualExogenousEmission[:, :, k]
                # Converts combined matrix to list and combines lists
                matlist = mat.tolist()
                #Combines the two lists
                combined_list = list(zip(lt, matlist))
                # Writes index specific parameter values to the text files
                f.write("\t[*,*,{0}]:\t{1}\t:=\n".format(y, column_string))
                for line in combined_list:
                    combinedflat = ''.join(str(line))
                    combinedflat = combinedflat.replace('[', '')
                    combinedflat = combinedflat.replace(']', '')
                    combinedflat = combinedflat.replace("'", '')
                    combinedflat = combinedflat.replace(",", '')
                    combinedflat = combinedflat.replace("(", '')
                    combinedflat = combinedflat.replace(")", '')
                    f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ModelPeriodExogenousEmission = np.zeros((lr,le))
            param = 'ModelPeriodExogenousEmission'  # Change this line
            f.write('#\n')
            columns = self.emission  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.ModelPeriodExogenousEmission[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')

            # ModelPeriodEmissionLimit = np.zeros((lr,le))
            param = 'ModelPeriodEmissionLimit'  # Change this line
            f.write('#\n')
            columns = self.emission  # Change this line
            column_string = ' '.join(columns)
            # Converts maxtrix rows to list
            array = np.array(self.region)  # Change this line
            array = array.T
            lt = array.tolist()
            # Creates 2D matrix for this value
            mat = self.ModelPeriodEmissionLimit[:, :]  # Change this line
            # Converts combined matrix to list and combines lists
            matlist = mat.tolist()
            #Combines the two lists
            combined_list = list(zip(lt, matlist))
            # Writes index specific parameter values to the text files
            f.write("param\t{0}\t:{1}:=\n".format(param, column_string))
            for line in combined_list:
                combinedflat = ''.join(str(line))
                combinedflat = combinedflat.replace('[', '')
                combinedflat = combinedflat.replace(']', '')
                combinedflat = combinedflat.replace("'", '')
                combinedflat = combinedflat.replace(",", '')
                combinedflat = combinedflat.replace("(", '')
                combinedflat = combinedflat.replace(")", '')
                f.write("{0}\n".format(combinedflat))
            f.write(';\n')
            f.write('end;\n')
            f.write('#')
        return



import os
import numpy as np
import pandas as pd
import inquirer as iq


class Forecasting:
    def __unit__(self):
        """Initialises the forecasting class
        """
        self.forecasts = None

    def energy_balance_base(self, root, IEA_World_Energy_Balances_1,
                            IEA_World_Energy_Balances_2,
                            create_excel_spreadsheet, output_file):
        """ Creates the baseline energy balance for forecasting

        Args:
            root (path): Path to provide access to all the files
            IEA_World_Energy_Balances_1 (str): File name for Energy Balance A to K
            IEA_World_Energy_Balances_2 ([type]): File name for Energy Balance L to Z
            create_excel_spreadsheet (boolean): True/false on whether to create a spreadsheet
            output_file (str): Name of output energy balance spreadsheet

        Returns:
            (dict): Dictionary of energy balances and unique lists (Use these key words to access: Energy Balances, Fuel, Geography, Technology)
        """
        IEAWEBAK = root / IEA_World_Energy_Balances_1
        IEAWEBLZ = root / IEA_World_Energy_Balances_2

        # Creates dataframes from IEA World Energy Statistics and Balances CSVs from Stats.OECD.org in the OECDiLibrary
        # Note the data is from #https:s//stats.oecd.org/ and #https://www-oecd-ilibrary-org.ezproxy.auckland.ac.nz/
        column_headers = [
            'ID', 'Unit', 'Geo_Code', 'Geo_Description', 'Prod_Code',
            'Prod_Description', 'Flow_Code', 'Flow_Description', 'Year',
            'Value(TJ)'
        ]
        f1 = open(IEAWEBAK, 'r')
        df_A = pd.read_csv(f1, header=None)
        df_A.columns = column_headers
        df_A.info(verbose=True)
        f2 = open(IEAWEBLZ, 'r')
        df_B = pd.read_csv(f2, header=None)
        df_B.columns = column_headers
        df_B.info(verbose=True)
        frames = [df_A, df_B]
        df = pd.concat(frames)
        df.info(verbose=True)

        # Closes the files
        f1.close()
        f2.close()

        # Finds the unique items in each list of the energy balance sheets
        unique_fuel = df.Prod_Description.unique()
        unique_geography = df.Geo_Description.unique()
        unique_technology = df.Flow_Description.unique()
        print(unique_geography)

        # Asks for a user to select a geography using the inquirer function
        selected_geo = input(
            "Please enter the geography you wish to extract energy balances:  "
        )

        # Creates a pivot table to display the data in the way similar to the Energy Balance Sheet (cols = Energy Product, rows = Energy Flows)
        energy_balance_pivot_table = pd.pivot_table(
            df,
            index=['Geo_Description', 'Flow_Description'],
            # Converted values to PJ
            values=['Value(TJ)'],
            columns=['Prod_Description'],
            aggfunc=[np.sum],
            fill_value=0)
        # Filters to the geography the user has selected
        Input_String = 'Geo_Description == ["' + selected_geo + '"]'
        geography_energy_balance_pivot_table = energy_balance_pivot_table.query(
            Input_String)

        if create_excel_spreadsheet == True:
            # Write the filtered pivot table to an excel file
            writer = pd.ExcelWriter(root / output_file)
            geography_energy_balance_pivot_table.to_excel(writer, selected_geo)
            writer.save()

        # Returns the unique lists and filtered pivot table as a dataframe
        return {
            "Energy Balances": geography_energy_balance_pivot_table,
            "Fuel": unique_fuel,
            "Geography": unique_geography,
            "Technology": unique_technology
        }

    def calculate_constant_average_growth_rate(self, start_year, end_year,
                                               start_value, end_value):
        """ Calculates the constant average growth rate (CAGR)

        Args:
            start_year (int): Starting year
            end_year (int): Ending year
            start_value (int): Initial value
            end_value (int): Final value

        Returns:
            cagr: Constant average growth rate (1+ decimal)
        """
        if start_value == 0 or (end_year - start_year) == 0:
            cagr = 1
        else:
            cagr = np.power((end_value / start_value),
                            (1 / (end_year - start_year)))
        return cagr

    def calculate_cagr_forecasts(self, cagr_dictionary, base_year_dictionary,
                                 fuel, year):
        """ Forecasts base year fuels by a constant average growth rate for a forecast period

        Args:
            cagr_dictionary (Dict): Dictionary of constant average growth rates per fuel
            base_year_dictionary ([type]): Dictionary of base year fuel consumption in energy types
            fuel (list): List of Fuels
            year (list): List of forecast years

        Returns:
            [float, array]: 2D Array of demand forecasts per fuel
        """
        # Initialises the size of the array
        forecast = np.ones((len(fuel), len(year)))

        # Set the first forecast as the base year
        for i in range(0, len(fuel), 1):
            forecast[i, 0] = base_year_dictionary[fuel[i]]

        # Calculates the forecasting
        for i in range(0, len(fuel), 1):
            for j in range(1, len(year), 1):
                forecast[i, j] = forecast[i, j - 1] * cagr_dictionary[fuel[j]]

        return forecast


from setuptools import setup
'Note: To upload new versions 1) cd to GOCPI 2) python setup.py sdist 3) twine upload dist/*'
'Note: To download 1) pip install --upgrade GOCPI'
'Note: Make your own python package: https://towardsdatascience.com/make-your-own-python-package-6d08a400fc2d'
'Note: PyPi https://pypi.org/manage/project/gocpi-functions/releases/#modal-close'
'Note: Enter dist/GOCPI-X.X.X.tar.gz to upload one file'

setup(name='GOCPI',
      version='3.0.916',
      description=
      'User Defined Functions and Energy Systems Data for GOCPI Project',
      packages=['GOCPI'],
      author_email='connormcdowall@gmail.com',
      zip_safe=False)

'/Users/connor/Google Drive/Documents/University/Courses/2020/ENGSCI 700A&B/GOCPI/src/GOCPI'


import os
import numpy as np
import pandas as pd


class CreateCases:
    """ A class of methods to create user-defined data cases
    """
    def __unit__(self):
        """ Sets the parameters and sets for the datacase
        """
        # Sets (placeholders for setting values)
        self.year = None
        self.region = None
        self.emission = None
        self.technology = None
        self.capacity_technology = None
        self.availability_technology = None
        self.fuel = None
        self.specified_fuel = None
        self.accumulated_fuel = None
        self.timeslice = None
        self.mode_of_operation = None
        self.storage = None
        self.daytype = None
        self.season = None
        self.dailytimebracket = None

        # Parameters
        self.Conversionls = None
        self.Conversionld = None
        self.Conversionlh = None
        self.DaysInDayType = None
        self.TradeRoute = None
        self.DepreciationMethod = None
        self.SpecifiedAnnualDemand = None
        self.SpecifiedDemandProfile = None
        self.AccumulatedAnnualDemand = None
        self.CapacityToActivityUnit = None
        self.CapacityFactor = None
        self.AvailabilityFactor = None
        self.OperationalLife = None
        self.ResidualCapacity = None
        self.InputActivityRatio = None
        self.OutputActivityRatio = None
        self.CapitalCost = None
        self.VariableCost = None
        self.FixedCost = None
        self.TechnologyToStorage = None
        self.TechnologyFromStorage = None
        self.StorageLevelStart = None
        self.StorageMaxChargeRate = None
        self.StorageMaxDischargeRate = None
        self.MinStorageCharge = None
        self.OperationalLifeStorage = None
        self.CapitalCostStorage = None
        self.ResidualStorageCapacity = None
        self.CapacityOfOneTechnologyUnit = None
        self.TotalAnnualMaxCapacity = None
        self.TotalAnnualMinCapacity = None
        self.TotalAnnualMaxCapacityInvestment = None
        self.TotalAnnualMinCapacityInvestment = None
        self.TotalTechnologyAnnualActivityLowerLimit = None
        self.TotalTechnologyAnnualActivityUpperLimit = None
        self.TotalTechnologyModelPeriodActivityUpperLimit = None
        self.TotalTechnologyModelPeriodActivityLowerLimit = None
        self.ReserveMarginTagTechnology = None
        self.ReserveMarginTagFuel = None
        self.ReserveMargin = None
        self.RETagTechnology = None
        self.RETagFuel = None
        self.REMinProductionTarget = None
        self.EmissionActivityRatio = None
        self.EmissionsPenalty = None
        self.AnnualExogenousEmission = None
        self.AnnualEmissionLimit = None
        self.ModelPeriodExogenousEmission = None
        self.ModelPeriodEmissionLimit = None

    def set_year(self, start_year, end_year, interval):
        """ Sets a list of forecast years

        Args:
            start_year (int): Starting year for forecasting (Less than end_year)
            end_year (int): Ending year for forecasting (Greater than start_year)
            interval (int): Gap for forecasting period
        """
        # Sets year array for new value
        year = []
        count = start_year
        while count <= end_year:
            year.append(str(count))
            count = count + interval
        self.year = year

    def set_region(self, regions):
        """ Sets the datacase's regions analysis

        Args:
            regions (list): list of regions 
        """
        self.region = regions

    def set_emission(self, emissions):
        """Sets the cases emission types

        Args:
            emissions (List): list of emission types
        """
        self.emission = emissions

    def set_technology(self, technology):
        """ Sets the cases technology type

        Args:
            technology (list): List of technologies
        """
        self.technology = technology

    def set_capacity_technology(self, capacity_technology):
        """ Sets the cases capacity_technology type

        Args:
            capacity_technology (list): List of technologies
        """
        self.capacity_technology = capacity_technology

    def set_availability_technology(self, availability_technology):
        """ Sets the cases availability_technology type

        Args:
            availability_technology (list): List of technologies
        """
        self.availability_technology = availability_technology

    def set_fuel(self, fuel):
        """ Sets the case's fuel types

        Args:
            fuel (list): list of fuels
        """
        self.fuel = fuel

    def set_specified_fuel(self, specified_fuel):
        """ Sets the case's specified fuel types

        Args:
            specified_fuel (list): list of specified fuels
        """
        self.specified_fuel = specified_fuel

    def set_accumulated_fuel(self, accumulated_fuel):
        """ Sets the case's accumulated fuel types

        Args:
            specified_fuel (list): list of specified fuels
        """
        self.accumulated_fuel = accumulated_fuel

    def set_timeslice(self, timeslice):
        """ Set of timeslices

        Args:
            timeslice (list): list of timeslices
        """
        self.timeslice = timeslice

    def set_mode_of_operation(self, num_modes_of_operation):
        """ Create the number of modes of operation (n = 1,...,num_modes_of_operation)

        Args:
            num_modes_of_operation (int): 
        """
        # Create set of mode_of_operation
        mode_of_operation = []
        count = 1
        while count <= num_modes_of_operation:
            mode_of_operation.append(str(count))
            count = count + 1
        self.mode_of_operation = mode_of_operation

    def set_storage(self, storage):
        """ Sets storage set of the datacase

        Args:
            storage (list): list of storage types
        """
        self.storage = storage

    def set_daytype(self, num_daytypes):
        """[summary]

        Args:
            num_daytypes (int): Number of daytypes
        """
        # Create set of daytypes
        daytype = []
        count = 1
        while count <= num_daytypes:
            daytype.append(str(count))
            count = count + 1
        self.daytype = daytype

    def set_season(self, num_seasons):
        """ Creates set of seasons

        Args:
            num_seasons (int): Number of seasons
        """
        # Create set of seasons
        season = []
        count = 1
        while count <= num_seasons:
            season.append(str(count))
            count = count + 1
        self.season = season

    def set_daily_time_bracket(self, num_dailytimebrackets):
        """ Creates set of daily time brackets

        Args:
            dailytimebracket (int): [description]
        """
        # Create set of dailytimebrackets
        dailytimebracket = []
        count = 1
        while count <= num_dailytimebrackets:
            dailytimebracket.append(str(count))
            count = count + 1
        self.dailytimebracket = dailytimebracket

    # Functions to define the parameters moving forward.
    def set_year_split(self, timeslices, years, splits):
        """ Creates 2D Numpy Array Parameter Splits.
            (Note: The index positions of timelices and splits must match)

        Args:
            timeslices (list): List of timeslices
            years (list): List of years
            splits (dict): A dictionary linking yearsplits to timeslices
        """
        # Creates a 2D YearSplit parameter
        YearSplit = np.ones((len(timeslices), len(years)))
        index = 0
        for time in timeslices:
            YearSplit[index, :] = splits[time]
            index = index + 1
        self.YearSplit = YearSplit

    def set_discount_rate(self, equity, debt, market_index,
                          cost_of_debt_pre_tax, risk_free_rate,
                          effective_tax_rate, preference_equity,
                          market_value_preference_shares, preference_dividends,
                          market_risk_coefficient):
        """[summary]

        Args:
            equity (dict): Dictionary of equity totals from treasury balance sheets
            debt (dict): Dictionary of equity totals from treasury balance sheets
            market_index (int, array): Regional monthly index returns (Arrays)
            cost_of_debt_pre_tax (dict): Dictionary of pre-tax cost of debts calculated from treasury balance sheets
            risk_free_rate (dict): Dictionary of risk free rates from 10 year swap rates for each region
            effective_tax_rate (dict): Dictionary of company tax rates for each region
            preference_equity (dict): Dictionary of preference equity for each region
            market_value_preference_shares (dict): Dictionary of the market value of prefence shares for each region
            preference_dividends (dict): Dictionary of prefence dividends for each region
            market_risk_coefficient (dict): Dictionary of markey risk co-efficients

        Returns:
            [int, array]: Numpy array of discount rates
        """
        # Creates empty dictionaries to stored values
        annualised_returns = {}
        cost_of_equity = {}
        cost_of_debt = {}
        cost_of_preference_equity = {}
        WACC = {}
        discount_rates = []
        # Calculates
        for region in market_index:
            # Calculates annualised returns for each regions market index
            annualised_rate_of_return = (np.power(
                (1 + ((market_index[region][-1] - market_index[region][0]) /
                      market_index[region][0])),
                (12 / len(market_index[region]))) - 1)
            annualised_returns[region] = annualised_rate_of_return
            # Calculates cost of equity
            cost_of_equity[region] = (
                risk_free_rate[region] + (market_risk_coefficient[region]) *
                (annualised_returns[region] - risk_free_rate[region]))
            # Calculates cost of debt
            cost_of_debt[region] = (cost_of_debt_pre_tax[region] / debt[region]
                                    ) * (1 - effective_tax_rate[region])
            # Calculates cost of preference equity
            cost_of_preference_equity[region] = preference_dividends[
                region] / market_value_preference_shares[region]
            # Calculates WACC
            WACC[region] = (
                cost_of_equity[region] *
                (equity[region] /
                 (equity[region] + debt[region] + preference_equity[region])) +
                cost_of_debt[region] *
                (debt[region] /
                 (equity[region] + debt[region] + preference_equity[region])) +
                cost_of_preference_equity[region] *
                (preference_equity[region] /
                 (equity[region] + debt[region] + preference_equity[region])))
            # Sets discount rates for each region
            discount_rates.append(WACC[region])
        # Set discount array
        self.DiscountRate = np.asarray(discount_rates)

    def set_day_split(self, daily_time_bracket, years, hour_split, num_days,
                      num_hours):
        """ Sets the day split parameter

        Args:
            daily_time_bracket (list): List of daily time brackets
            years (list): List of year
            hour_split (dict): Dictonary of hours in a daily time bracket 
            num_days (int): Number of days in a year
            num_hours (int): Number of hours in a day
        """
        # Initilises the DaySplit Array
        DaySplit = np.ones((len(daily_time_bracket), len(years)))
        index = 0
        for split in daily_time_bracket:
            DaySplit[index, :] = hour_split[split] / (num_days * num_hours)
            index = index + 1
        self.DaySplit = DaySplit

    def set_conversion_ls(self, timeslice, season, link):
        """ Sets the Conversionls parameter

        Args:
            timeslice (list): List of timeslices
            season (list): List of seasons
            link (dict): Dictionary describing the connection between timeslices and seasons
        """
        Conversionls = np.zeros((len(timeslice), len(season)))
        for i in range(0, len(timeslice), 1):
            for j in range(0, len(season), 1):
                if link[timeslice[i]] == season[j]:
                    Conversionls[i, j] = 1

        self.Conversionls = Conversionls

    def set_conversion_ld(self, timeslice, daytype, link):
        """ Sets the Conversionld parameter

        Args:
            timeslice (list): List of timeslices
            daytype (list): List of daytypes
            link (dict): Dictionary describing the connection between timeslices and daytypes
        """
        Conversionld = np.zeros((len(timeslice), len(daytype)))
        for i in range(0, len(timeslice), 1):
            Conversionld[i, :] = link[timeslice[i]]

        self.Conversionld = Conversionld

    def set_conversion_lh(self, timeslice, dailytimebracket, link, override):
        """ Sets the Conversionlh parameter

        Args:
            timeslice (list): List of timeslices
            dailytimebracket (list): List of dailytimebracket
            link (dict): Dictionary describing the connection between timeslices and dailytimebrackets
            override (int, array): Override if want to manually put in the array
        """
        if override == None:
            Conversionlh = np.zeros((len(timeslice), len(dailytimebracket)))
            for i in range(0, len(timeslice), 1):
                Conversionlh[i, :] = link[timeslice[i]]
            self.Conversionlh = Conversionlh
        else:
            self.Conversionlh = override

    def set_days_in_day_type(self, season, daytype, year, link, override):
        """ Sets the DaysInDayType parameter

        Args:
            season (list): List of seasons
            daytype (list): List of daytypes
            year (list): List of years
            link (dict): Dictionary relating seasons to daytypes
            override (int, array): Override if want to manually put in the array
        """
        if override == None:
            DaysInDayType = np.zeros((len(season), len(daytype), len(year)))
            for i in range(0, len(season), 1):
                for j in range(0, len(year), 1):
                    DaysInDayType[i, :, j] = link[season[i]]
            self.DaysInDayType = DaysInDayType
        else:
            self.DaysInDayType = override

    def set_trade_route(self, trade):
        """ Sets the TradeRoute parameter between regions
            (Assume it is the same across fuels and years)

        Args:
            trade (int ,array): 4D array representing trade relationships 
                                between regions, fuels and years. You 
                                must model this manually.
        """
        self.TradeRoute = trade

    def set_depreciation_method(self, region, methods, override):
        """ Sets DepreciationMethod
            (1 = Sinking Fund Depreciation, 2 = Straightline Depreciation)

        Args:
            region (list): List of regions
            override (int, array): Manual array for setting depreciation methods
            methods (dict): Dictionary assigning methods to regions
        """

        if override == None:
            depreciation_method = np.ones((len(region)))
            for i in range(0, len(region), 1):
                depreciation_method[i] = methods[region[i]]
            self.DepreciationMethod = depreciation_method
        else:
            self.DepreciationMethod = override

    def set_specified_annual_demand(self, specified_forecast):
        """ Sets the annual demand for fuels per region over the forecast period (Must be accurate)

        Args:
            forecast (float, array): The forecast array of size (len(region),len(fuel),len(year))
        """
        self.SpecifiedAnnualDemand = specified_forecast

    def set_specified_demand_profile(self, specified_annual_demand, region,
                                     fuel, year, timeslice, profile, override):
        """ Sets the specified annual demand profiles using the specified annual demand.

        Args:
            specified_annual_demand (float, array): Specified annual demand profiles
            region (list): List of regions
            fuel (list): List of fuels
            year (list): List of years
            timeslice (list): List of timeslices
            profile (Dict): Dictionary of fuel allocations to timeslices
            override (float, array): Manual override for the specified annual demand profiles.
        """
        # Initialises the linear array
        demand_profile = np.zeros(
            (len(region), len(fuel), len(timeslice), len(year)))
        if override == None:
            # Calculates the demand profile
            for place in region:
                for fuel_type in fuel:
                    for time in timeslice:
                        for year_num in year:
                            region_index = region.index(place)
                            fuel_index = fuel.index(fuel_type)
                            timeslice_index = timeslice.index(time)
                            year_index = year.index(year_num)
                            demand_profile[region_index, fuel_index,
                                           timeslice_index,
                                           year_index] = profile[time]

            self.SpecifiedDemandProfile = demand_profile
        else:
            self.SpecifiedDemandProfile = override

    def set_accumulated_annual_demand(self, accumulated_forecast):
        """ Sets the accumulated annual demand for fuels per region over the forecast period.
            This function relies on a similar forecasting methodology as set_specific_demand.
            Fuels set in this function cannot be defined in set_specific_demand.

        Args:
            accumulated_forecast (float, array): The forecast array of size (len(region),len(fuel),len(year))
        """
        self.AccumulatedAnnualDemand = accumulated_forecast

    def set_capacity_to_activity_unit(self, region, technology,
                                      capacity_dictionaries, override):
        """ Sets the capacity to activity parameter

        Args:
            region (list): List of regions
            technology (list): List of technologies
            capacity_dictionaries (list): List of dictionaries to assign value
            override (float, array) = 2D Array to assign override values
        """
        if override == None:
            cap_to_act = np.zeros((len(region), len(technology)))
            for i in range(0, len(capacity_dictionaries), 1):
                for j in range(0, len(technology), 1):
                    cap_to_act[i, j] = capacity_dictionaries[i][technology[j]]
            self.CapacityToActivityUnit = cap_to_act
        else:
            self.CapacityToActivityUnit = override

    def set_capacity_factor(self, factor_matrix):
        """ Sets capacity factors for conversion technologies.

        Args:
            factor_matrix (float, array); Capacity Factors
        """
        self.CapacityFactor = factor_matrix

    def set_availability_factor(self, availablility_matrix):
        """ Sets the availability factors

        Args:
            availablility_matrix (float, array): Matrix describing availability factors for given technologies
        """
        self.AvailabilityFactor = availablility_matrix

    def set_operational_life(self, operational_lives):
        """ Sets operational life

        Args:
            operational_lives (list):
        """

    def set_residual_capacity(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_input_activity_ratio(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_output_activity_ratio(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_capital_cost(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_variable_cost(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_fixed_cost(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_technology_to_storage(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_technology_from_storage(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_min_storage_charge(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_operational_life_storage(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_capital_cost_storage(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_storage_level_start(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_storage_max_charge_rate(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_storage_max_discharge_rate(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_residual_storage_capacity(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """
        self.ResidualStorageCapacity = None

    def set_capacity_of_one_technology_unit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_annual_max_capacity(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_annual_min_capacity(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_technology_annual_activity_lower_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_technology_annual_activity_upper_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_technology_period_activity_upper_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_total_technology_period_activity_lower_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_reserve_margin_tag_technology(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_reserve_margin_tag_fuel(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_reserve_margin(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_re_tag_technology(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_re_tag_fuel(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_re_min_production_target(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_emission_activity_ratio(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_emissions_penalty(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_annual_exogenous_emission(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_annual_emission_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_model_period_exogenous_emission(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """

    def set_model_period_emission_limit(self, parameters):
        """[summary]

        Args:
            parameters ([type]): [description]
        """